(*
 * Wes Weimer
 * Sun May 16 22:17:03 PST 2004
 *
 * Java Abstract Syntax Tree
 *
 * In general, Java AST concepts use the same names as: 
 *
 * http://java.sun.com/docs/books/jls/second_edition/html/jTOC.doc.html
 *
 *)

(* An identifier like "mySocket" *) 
type ident = string

(* "java.net.Socket" or "localVar.someField" *) 
type qualified_ident = ident list 

(* A "foo node" is both a "foo" and some location information. We currently
 * keep around location information for expressions and statements. *) 
type 'a node = {
  mutable node : 'a;
          span : Jabsutil.span; 
} 

(* One "foo.java" file yields one compilation unit when parsed 
 * Use "Jabsparse.parse_file filename" (or somesuch) to get one of these. *) 
type compilation_unit = {
          cu_filename : string;
          package : qualified_ident option ; 
          imports : import list ;
  mutable decls : type_decl list ;
} 

and import = 
  SingleTypeImport of qualified_ident (* import java.io.IOException *) 
| TypeImportOnDemand of qualified_ident (* import java.awt.* *) 

and class_decl = {
          c_decl_modifiers : modifier list ;
          c_name : ident ;
          c_extends : typ option ;
          c_implements : typ list;
          c_body : class_body ;
} 

and interface_decl = {
          i_decl_modifiers : modifier list ;
          i_name : ident ; 
          i_extends : typ list;
          i_body : interface_body ; 
} 

and modifier = Static | Abstract | Final | Public | Protected | Private 
  | Strictfp | Native | Synchronized | Transient | Volatile

and type_decl = 
  | ClassDecl of class_decl 
  | InterfaceDecl of interface_decl  

and class_body = class_body_decl list  

and class_body_decl = 
  | ClassMemberDecl of class_member_decl
  | InstanceInitializer of block
  | StaticInitializer of block
  | ConstructorDecl of constructor_decl 

and class_member_decl = 
  | FieldDecl of field_decl
  | MethodDecl of method_decl
  | ClassOrInterfaceDecl of type_decl 
  | NopClassMemberDecl  

and interface_body = class_member_decl list 

and field_decl = {
  fd_modifiers : modifier list;
  fd_type : typ ; 
  fd_vd : variable_declarator list; 
} 

and variable_declarator = {
  vd_name : ident ;
  vd_brackets : int ; 
  vd_init : variable_initializer option ; 
} 

and variable_initializer = 
  | VI_ExpInit of expn  
  | VI_ArrayInit of array_initializer

and array_initializer = variable_initializer list 

and typ = Primitive of primitive_type
        | Reference of reference_type

and primitive_type = Boolean | Char | Byte | Short | Int | Long 
  | Float | Double 

and reference_type = 
        | ClassOrInterfaceType of qualified_ident
        | ArrayType of typ 

and type_or_void =  (* method result type *) 
  | Type of typ
  | Void

and constructor_decl = {
  cons_modifiers : modifier list;
  cons_name      : ident ;
  cons_formals   : formal list; 
  cons_throws    : qualified_ident list ; 
  cons_body      : block ; 
} 

and method_decl = {
  md_modifiers : modifier list; 
  md_type : type_or_void ;
  md_name : ident ; 
  md_formals : formal list;
  md_throws : qualified_ident list ;
  md_body : block option ; 
} 

and formal = {
  f_final : bool ;
  f_type : typ ;
  f_name : ident ;
  f_brackets : int ;
} 

and block = block_statement list

and block_statement = 
  | LocalVariableDecl of local_variable_decl
  | BlockClassDecl of class_decl 
  | Statement of stmtn

and local_variable_decl = {
  lv_final : bool ;
  lv_type  : typ ;
  lv_vd : variable_declarator list ; 
} 

and stmtn = stmt node 

and stmt = 
  | Labeled of ident * stmtn 
  | IfThen of expn * stmtn
  | IfThenElse of expn * stmtn * stmtn
  | ExpStmt of exp_stmt
  | Switch of expn * switch_block
  | While of expn * stmtn
  | Do of stmtn * expn
  | For of for_stmt
  | BlockStmt of block 
  | EmptyStmt 
  | Break of ident option
  | Continue of ident option
  | Return of expn option
  | Throw of expn 
  | SynchronizedExp of expn * block
  | Try of try_stmt

and try_stmt = { 
  try_block : block ;
  try_catch : catches ; 
  try_finally : block option ; 
} 

and catches = (formal * block) list 

and for_stmt = {
  for_init : for_init option ;
  for_guard : expn option ;
  for_update : esn list;
  for_body : stmtn ;
} 

and for_init = 
  | ForInitStmtExp of esn list
  | ForInitVarDecl of local_variable_decl 

and switch_block = {
  sb_stmt_groups : switch_block_stmt_group list ;
} 

and switch_block_stmt_group = {
  sbsg_labels : switch_label list ;
  sbsg_stmts  : block_statement list ; 
} 

and switch_label = 
  | Default
  | Case of expn

and esn = exp_stmt node 

and exp_stmt = 
  | Assignment of aop * expn * expn
  | PreOp of preop * expn 
  | PostOp of expn * postop  
  | MethodInvoc of ident * (expn list)  
  | MethodInvocExp of expn * ident * (expn list)
  | MethodInvocSuper of (ident option) * (expn list)
  | MethodInvocThis of (expn list)
  | MethodInvocClassSuper of qualified_ident * ident * (expn list)
  | ClassInstanceCreationType of typ * class_instance_creation
  | ClassInstanceCreationExp of expn * ident * class_instance_creation

and aop = Assign | PlusAssign | MinusAssign | StarAssign | SlashAssign 
  | AndAssign | OrAssign | XorAssign | RemAssign | LShiftAssign 
  | RSignedShiftAssign | RUnsignedShiftAssign 

and preop = PreInc | PreDec | Bang | Tilde | PrePlus | PreMinus 

and postop = PostInc | PostDec

and class_instance_creation = {
  cic_args    : expn list;
  cic_body    : class_body option ;
} 

and expn = exp node 

and exp = 
  | Literal of literal 
  | TypeDotClass of typ
  | VoidDotClass
  | This
  | ClassNameDotThis of qualified_ident
  | ES of exp_stmt  (* Expression Statement *) 
  | ArrayCreation of array_creation
  | ArrayCreationInit of array_creation_init
  | Field of expn * ident 
  | SuperField of ident 
  | ClassSuperField of qualified_ident * ident
  | ArrayAccess of expn * expn (* a[b] *) 
  | Cast of typ * expn 
  | BinOp of bop * expn * expn 
  | InstanceOf of expn * typ 
  | Conditional of expn * expn * expn
  | ExpressionName of qualified_ident 

and bop = SC_Or | SC_And | Bit_Or | Bit_Xor | Bit_And | Eq | Ne | Gt | Lt | Ge 
  | Le | LShift | RSignedShift | RUnsignedShift | Plus | Minus | Star 
  | Slash | Rem  

and array_creation = {
  ac_type     : typ ;
  ac_dim_expr : expn list ;
  ac_dims     : int ;
} 

and array_creation_init = {
  aci_type     : typ ;
  aci_dims     : int ;
  aci_init     : array_initializer;
} 

and literal =
    IntLit of string
  | FloatLit of string
  | CharLit of string
  | StringLit of string
  | True 
  | False
  | Null
