(* Pretty-Printer for Jabs that emits valid Java code *)
open Jabs
open Pretty
(* Parentheses level. An expression "a op b" is printed parenthesized if its 
 * parentheses level is >= that that of its context. Identifiers have the 
 * lowest level and weakly binding operators (e.g. |) have the largest level. 
 * The correctness criterion is that a smaller level MUST correspond to a 
 * stronger precedence!
 *)
let getParenthLevel = function 
  | ES(Assignment(_,_,_)) 
  | Conditional(_,_,_) ->       99
  | BinOp((SC_Or),_,_) ->       98
  | BinOp((SC_And),_,_) ->      97
  | BinOp((Bit_Or),_,_) ->      96
  | BinOp((Bit_Xor),_,_) ->     95
  | BinOp((Bit_And),_,_) ->     94
  | BinOp((Eq|Ne),_,_) ->       93

  | BinOp((Lt|Gt|Le|Ge),_,_) 
  | InstanceOf(_,_) ->          92

  | BinOp((LShift|RSignedShift|RUnsignedShift),_,_) -> 91
  | BinOp((Minus),_,_) ->  90
  | BinOp((Plus),_,_) ->  89
  | BinOp((Star|Slash|Rem),_,_) ->  88
  | ES(PreOp((PreInc|PreDec|PrePlus|PreMinus),_)) ->  87

  | Cast(_,_) 
  | ES(PreOp((Bang|Tilde),_)) 
  | ES(PostOp(_,(PostInc|PostDec))) ->  86

  | ExpressionName(_) 
  | Literal(_)
  | TypeDotClass(_)
  | VoidDotClass
  | This
  | ClassNameDotThis(_) 
  | ES(ClassInstanceCreationExp(_,_,_))
  | ES(ClassInstanceCreationType(_,_))
  | Field(_,_)
  | SuperField(_)
  | ClassSuperField(_)
  | ES(MethodInvoc(_,_)) 
  | ES(MethodInvocExp(_,_,_)) 
  | ES(MethodInvocSuper(_,_)) 
  | ES(MethodInvocThis(_)) 
  | ES(MethodInvocClassSuper(_)) 
  | ArrayAccess(_,_)  
  | ArrayCreation(_) 
  | ArrayCreationInit(_) ->        85

let d_i () i = text i 

let d_qi () q = (docList (chr '.') (fun s -> d_i () s)) () q

let rec d_cu () cu = 
  dprintf "/* jabs2java: %s */@!" cu.cu_filename
  ++ (match cu.package with
      | None -> nil
      | Some(qi) -> dprintf "package %a;@!" d_qi qi)
  ++ (docList (nil) (fun i -> match i with
      | SingleTypeImport(qi) -> dprintf "import %a;@!" d_qi qi
      | TypeImportOnDemand(qi) -> dprintf "import %a.*;@!" d_qi qi)) ()
      cu.imports 
  ++ (docList (nil) (d_type_decl ())) () cu.decls

and d_type_decl () td = match td with
  | ClassDecl(cd) -> d_class_decl () cd
  | InterfaceDecl(i) -> d_interface_decl () i

and d_modifier () (a12 : modifier) =
  begin match a12 with
    Static -> text "static"
  | Abstract -> text "abstract"
  | Final -> text "final"
  | Public -> text "public"
  | Protected -> text "protected"
  | Private -> text "private"
  | Strictfp -> text "strictfp"
  | Native -> text "native"
  | Synchronized -> text "synchronized"
  | Transient -> text "transient"
  | Volatile -> text "volatile"
  end

and d_modifiers () ml = 
  (docList (nil) (fun m -> d_modifier () m ++ break)) () ml

and d_class_decl () cd = 
  dprintf "%aclass %a" d_modifiers cd.c_decl_modifiers
    d_i cd.c_name 
  ++ (match cd.c_extends with
      | None -> nil
      | Some(t) -> dprintf "@?extends %a" d_type t)
  ++ (match cd.c_implements with
      | [] -> nil
      | lst -> dprintf "@?implements@?%a" 
        (docList (chr ',' ++ break) (d_type ())) lst)
  ++ text " {" ++ line ++ text "  " ++ align
  ++ d_class_body () cd.c_body
  ++ unalign ++ text "}" ++ line

and d_interface_decl () id = 
  dprintf "%ainterface %a" d_modifiers id.i_decl_modifiers
    d_i id.i_name 
  ++ (match id.i_extends with
      | [] -> nil
      | lst -> dprintf "@?extends %a" 
            (docList (chr ',' ++ break) (d_type ())) lst)
  ++ text " {" ++ line ++ text "  " ++ align
  ++ d_interface_body () id.i_body
  ++ unalign ++ text "}" ++ line

and d_class_body () cb = 
  (docList (nil) (fun cbd -> d_class_body_decl () cbd ++ line)) () cb

and d_interface_body () ib =
  (docList (nil) (fun cbd -> d_class_member_decl () cbd ++ line)) () ib

and d_class_body_decl () cbd = match cbd with
  | ClassMemberDecl(cmd) -> d_class_member_decl () cmd 
  | InstanceInitializer(b) -> dprintf "{ @[%a@]@!}@!" d_block b
  | StaticInitializer(b) -> dprintf "static { @[%a@]@!}@!" d_block b
  | ConstructorDecl(cd) -> d_constructor_decl () cd

and d_type () t = match t with
  | Primitive(pt) -> d_primitive_type () pt
  | Reference(rt) -> d_reference_type () rt

and d_primitive_type () (a36 : primitive_type) =
  begin match a36 with
    Boolean -> text "boolean"
  | Char -> text "char"
  | Byte -> text "byte"
  | Short -> text "short"
  | Int -> text "int"
  | Long -> text "long"
  | Float -> text "float"
  | Double -> text "double"
  end

and d_reference_type () rt = match rt with
  | ClassOrInterfaceType(qi) -> d_qi () qi
  | ArrayType(t) -> d_type () t ++ text "[]" 

and d_type_or_void () tv = match tv with
  | Type(t) -> d_type () t
  | Void -> text "void" 

and d_class_member_decl () cmd = match cmd with
  | FieldDecl(f) -> d_field_decl () f  
  | MethodDecl(m) -> d_method_decl () m 
  | ClassOrInterfaceDecl(td) -> d_type_decl () td
  | NopClassMemberDecl -> text ";" 

and d_field_decl () f = 
  dprintf "%a%a %a;"
    d_modifiers f.fd_modifiers 
    d_type f.fd_type 
    (docList (chr ',' ++ break) (d_variable_declarator ())) f.fd_vd

and d_method_decl () md = 
  dprintf "%a%a %a(%a)" 
    d_modifiers md.md_modifiers
    d_type_or_void md.md_type
    d_i md.md_name 
    d_formals md.md_formals  
  ++ (match md.md_throws with
      | [] -> nil
      | qil -> dprintf " throws@?%a" (docList break (d_qi ())) qil)
  ++ break ++ (match md.md_body with
      | None -> text ";" 
      | Some(b) -> d_block () b)

and d_constructor_decl () cd = 
  dprintf "%a%a(%a)" 
    d_modifiers cd.cons_modifiers
    d_i cd.cons_name 
    d_formals cd.cons_formals  
  ++ (match cd.cons_throws with
      | [] -> nil
      | qil -> dprintf " throws@?%a" (docList break (d_qi ())) qil)
  ++ break ++ d_block () cd.cons_body

and d_formals () fl = docList (chr ',' ++ break) (d_formal ()) () fl 

and d_brackets () i = match i with
  | 0 -> nil
  | j -> text "[]" ++ d_brackets () (j - 1)

and d_formal () f = 
  (if f.f_final then text "final" ++ break else nil) 
  ++ d_type () f.f_type ++ break
  ++ d_i () f.f_name 
  ++ d_brackets () f.f_brackets 

and d_block () b =  
  match b with 
  | [] -> text "{ }" 
  | lst -> dprintf "{@!  @[%a@]@!}" d_block_stmt_list lst 

and d_block_stmt_list () bsl = docList (line) (d_block_stmt ()) () bsl 

and d_block_stmt () bs = match bs with
  | LocalVariableDecl(lvd) -> d_local_variable_decl () lvd ++ text ";" 
  | BlockClassDecl(cd) -> d_class_decl () cd 
  | Statement(s) -> d_stmtn () s 

and d_local_variable_decl () lvd = 
  (if lvd.lv_final then text "final" ++ break else nil) 
  ++ d_type () lvd.lv_type ++ break 
  ++ docList (chr ',' ++ break) (d_variable_declarator ()) () lvd.lv_vd

and d_variable_declarator () vd = 
  d_i () vd.vd_name 
  ++ d_brackets () vd.vd_brackets
  ++ (match vd.vd_init with
      | None -> nil
      | Some(i) -> dprintf "@?=@?%a" d_variable_initializer i)

and d_variable_initializer () vi = match vi with
  | VI_ExpInit(e) -> d_expn () e
  | VI_ArrayInit(ai) -> dprintf "{@[%a@]}" d_array_initializer ai

and d_array_initializer () ai = 
  docList (chr ',' ++ break) (d_variable_initializer ()) () ai 

and d_stmtn () sn = d_stmt () sn.node

and d_stmt () s = match s with
  | Labeled(i,s) -> dprintf "%a: %a" d_i i d_stmtn s
  | IfThen(e,s) -> dprintf "if@?(@[%a@])@?%a" d_expn e d_stmtn s
  | IfThenElse(e,a,b) -> 
      dprintf "if@?(@[%a@])@?%a@?else@?%a" d_expn e d_stmtn a d_stmtn b
  | ExpStmt(e) -> dprintf "%a;" d_exp_stmt e
  | Switch(e,sb) -> dprintf "switch@?(@[%a@])@?%a"
      d_expn e d_switch_block sb 
  | While(e,s) -> dprintf "while@[@?(@[%a@])@?%a@]" d_expn e d_stmtn s
  | Do(s,e) -> dprintf "do@?@[%a@]@?while@?(@[%a@]);" d_stmtn s d_expn e
  | For(f) -> d_for () f 
  | BlockStmt(b) -> d_block () b 
  | EmptyStmt -> text ";" 
  | Break(None) -> text "break;" 
  | Break(Some(i)) -> dprintf "break %a;" d_i i 
  | Continue(None) -> text "continue;" 
  | Continue(Some(i)) -> dprintf "continue %a;" d_i i 
  | Return(None) -> text "return;" 
  | Return(Some(e)) -> dprintf "return %a;" d_expn e 
  | Throw(e) -> dprintf "throw @[%a@];" d_expn e
  | SynchronizedExp(e,b) -> dprintf "synchronized@?(@[%a@])@?%a"
      d_expn e d_block b
  | Try(t) -> d_try () t  

and d_args () args = 
  (docList (chr ',' ++ break) (d_expn ())) () args 

and d_class_body_opt () cbo = match cbo with
  | None -> nil
  | Some(cb) -> text "{" ++ d_class_body () cb ++ text "}" 

and d_esn () esn = d_exp_stmt () esn.node

and d_exp_stmt () es = match es with
  | Assignment(aop,e1,e2) -> dprintf "@[%a@]@?%a@?@[%a@]" 
      d_expn e1 d_aop aop d_expn e2
  | PreOp(p,e) -> dprintf "%a@[%a@]" d_preop p d_expn e
  | PostOp(e,p) -> dprintf "@[%a@]%a" d_expn e d_postop p 
  | MethodInvoc(i,args) -> dprintf "%a@?(@[%a@])@?" d_i i d_args args
  | MethodInvocExp(e,i,args) -> dprintf "@[%a@].%a@?(@[%a@])@?" 
      d_expn e d_i i d_args args
  | MethodInvocSuper(None,args) -> dprintf "super@?(@[%a@])@?" d_args args
  | MethodInvocThis(args) -> dprintf "this(@[%a@])" d_args args
  | MethodInvocSuper(Some(i),args) -> dprintf "super.%a@?(@[%a@])@?" 
      d_i i d_args args
  | MethodInvocClassSuper(qi,i,args) -> dprintf "%a.super.%a@?(@[%a@])@?" 
      d_qi qi d_i i d_args args 
  | ClassInstanceCreationType(t,cic) ->
      dprintf "new@?%a@?(@[%a@])@?%a"
        d_type t d_args cic.cic_args d_class_body_opt cic.cic_body
  | ClassInstanceCreationExp(e,i,cic) ->
      dprintf "@[%a@].new@?%a@?(@[%a@])@?%a"
        d_expn e d_i i d_args cic.cic_args d_class_body_opt cic.cic_body

and d_expPrec contextPrec () (e : exp) = 
  let thisLevel = getParenthLevel e in
  let needParens = 
    if thisLevel > contextPrec then true
    else false
  in 
  if needParens then
    chr '(' ++ d_exp () e ++ chr ')'
  else
    d_exp () e

and d_exp_stmtPrec (contextPrec:int) () (es : exp_stmt) = 
  let thisLevel = getParenthLevel (ES(es)) in
  let needParens = 
    if thisLevel >= contextPrec then true
    else false
  in 
  if needParens then
    chr '(' ++ d_exp_stmt () es ++ chr ')'
  else
    d_exp_stmt () es

and d_expn () en = d_exp () en.node 

and d_exp () e = 
  let level = getParenthLevel e in
  let d_subexp () e = d_expPrec level () e.node in 
  (match e with
  | Literal(l) -> d_literal () l
  | TypeDotClass(t) -> dprintf "%a.class" d_type t
  | VoidDotClass -> text "void.class" 
  | This -> text "this" 
  | ClassNameDotThis(qi) -> dprintf "%a.this" d_qi qi 
  | ES(es) -> d_exp_stmt () es 
  | ArrayCreation(ac) -> dprintf "new@?%a%a%a" d_type ac.ac_type
        d_dim_expr ac.ac_dim_expr d_brackets ac.ac_dims
  | ArrayCreationInit(aci) -> 
      dprintf "new@?%a%a@?{%a}"
        d_type aci.aci_type 
        d_brackets aci.aci_dims 
        d_array_initializer aci.aci_init
  | Field(e,i) -> dprintf "%a.%a" d_subexp e d_i i
  | SuperField(i) -> dprintf "super.%a" d_i i 
  | ClassSuperField(qi,i) -> dprintf "%a.super.%a" d_qi qi d_i i
  | ArrayAccess(a,b) -> dprintf "%a[%a]" d_subexp a d_subexp b
  | Cast(t,e) -> dprintf "((@[%a@])@?%a)" d_type t d_subexp e
  | BinOp(bop,a,b) -> dprintf "%a %a %a" d_subexp a d_bop bop d_subexp b 
  | InstanceOf(e,t) -> dprintf "(%a instanceof %a)" d_subexp e d_type t
  | Conditional(a,b,c) -> dprintf "%a ? %a : %a" d_subexp a d_subexp b 
      d_subexp c
  | ExpressionName(qi) -> d_qi () qi ) 

and d_dim_expr () el = 
  docList (nil) (fun e -> dprintf "[%a]" d_expn e) () el 

and d_switch_block () sb = 
  dprintf "{@!  @[%a@]" 
    d_sbsg_list sb.sb_stmt_groups 
    (*
  ++ (if (sb.sb_extra_labels = []) then nil 
      else dprintf "@!  @[%a@]" d_label_list sb.sb_extra_labels)
      *)
  ++ line ++ text "}"

and d_label_list () ll = 
  (docList (line) (fun l -> d_label () l)) () ll 

and d_label () l = match l with
  | Default -> text "default:" ++ break 
  | Case(e) -> dprintf "case @[%a@]:" d_expn e ++ break 

and d_sbsg_list () sbsgl = 
  (docList (line) (fun sbsg -> d_sbsg () sbsg)) () sbsgl 

and d_sbsg () sbsg = 
  dprintf "%a@[%a@]" d_label_list sbsg.sbsg_labels
                 d_block_stmt_list sbsg.sbsg_stmts 

and d_for () f = 
  dprintf "for@?(@[%a@?;@?%a@?;@?%a@])@?%a"
    d_for_init_opt f.for_init
    d_exp_opt f.for_guard
    d_for_update f.for_update
    d_stmtn f.for_body

and d_exp_opt () eo = match eo with
  | None -> nil
  | Some(e) -> d_expn () e

and d_for_init_opt () fio = match fio with
  | None -> nil
  | Some(fi) -> begin match fi with 
    | ForInitStmtExp(esl) -> docList (chr ',' ++ break) 
        (d_esn ()) () esl 
    | ForInitVarDecl(lvd) -> d_local_variable_decl () lvd 
  end 

and d_for_update () esl = docList (chr ',' ++ break)
      (d_esn ()) () esl

and d_try () t = 
  dprintf "try %a" d_block t.try_block 
  ++ docList (nil) (fun (f,b) -> 
    dprintf "@?catch (@[%a@]) %a" d_formal f d_block b) () t.try_catch
  ++ (match t.try_finally with
    | None -> nil
    | Some(b) -> dprintf "@?finally %a" d_block b)

and d_aop () (a111 : aop) =
  begin match a111 with
    Assign -> text "="
  | PlusAssign -> text "+="
  | MinusAssign -> text "-="
  | StarAssign -> text "*="
  | SlashAssign -> text "/="
  | AndAssign -> text "&="
  | OrAssign -> text "|="
  | XorAssign -> text "^="
  | RemAssign -> text "%="
  | LShiftAssign -> text "<<="
  | RSignedShiftAssign -> text ">>="
  | RUnsignedShiftAssign -> text ">>>="
  end

and d_bop () (a140 : bop) =
  begin match a140 with
    SC_Or -> text "||"
  | SC_And -> text "&&"
  | Bit_Or -> text "|"
  | Bit_Xor -> text "^"
  | Bit_And -> text "&"
  | Eq -> text "=="
  | Ne -> text "!="
  | Gt -> text ">"
  | Lt -> text "<"
  | Ge -> text ">="
  | Le -> text "<="
  | LShift -> text "<<"
  | RSignedShift -> text ">>"
  | RUnsignedShift -> text ">>>"
  | Plus -> text "+"
  | Minus -> text "-"
  | Star -> text "*"
  | Slash -> text "/"
  | Rem -> text "%"
  end

and d_preop () (a112 : preop) =
  begin match a112 with
    PreInc -> text "++"
  | PreDec -> text "--"
  | Bang -> text "!"
  | Tilde -> text "~"
  | PrePlus -> text "+"
  | PreMinus -> text "-"
  end

and d_postop () (a113 : postop) =
  begin match a113 with
    PostInc -> text "++"
  | PostDec -> text "--"
  end

and d_literal () l = match l with
  | IntLit(s) 
  | FloatLit(s) -> text s 
  | CharLit(s) -> dprintf "'%s'" s
  | StringLit(s) -> dprintf "\"%s\"" s
  | True -> text "true" 
  | False -> text "false" 
  | Null -> text "null"
