open Pretty

let version = 1 

type loc = {
  l_line   : int ;  (* line in file *) 
  l_col    : int ;  (* column in line *) 
  l_offset : int ;  (* byte from beginning of file *) 
} 

type span = loc * loc 

let loc_unknown = { l_line = 0; l_col = 0; l_offset = 0; }

let span_unknown = (loc_unknown,loc_unknown)

let loc_to_str l = 
  Printf.sprintf "[Line %d, Col %d, Off %d]" l.l_line l.l_col l.l_offset

let span_to_str (l1,l2) = 
  Printf.sprintf "[Line %4d, Col %2d -- Line %4d, Col %2d]" 
    l1.l_line l1.l_col
    l2.l_line l2.l_col

let d_span () (l1,l2) = 
  dprintf "[@[Line %d, Col %d@?-- Line %d, Col %d@]]" 
    l1.l_line l1.l_col
    l2.l_line l2.l_col 

(* for convenient parse errors *)

type input_context = {
  mutable loc : loc ;
  mutable filename : string ;
  mutable lexbuf : Lexing.lexbuf ; 
  mutable warn_only : bool ; 
} 
let the_context = {
  loc = loc_unknown ;
  filename = "" ;
  lexbuf = Lexing.from_string "";
  warn_only = false ;
} 

let lex_init (filename: string) (inchannel: in_channel) : Lexing.lexbuf =
  let lexbuf = Lexing.from_channel inchannel in
  the_context.loc <- { l_line = 1; l_col = 1; l_offset = 1; } ;
  the_context.filename <- filename ;
  the_context.lexbuf <- lexbuf ;
  lexbuf 

let newline () = 
  let c = the_context in 
  c.loc <- { l_line = c.loc.l_line + 1 ; 
             l_col = 1 ; 
             l_offset = c.loc.l_offset + 1 ; } 

let tab () = 
  let c = the_context in
  c.loc <- { c.loc with l_col = c.loc.l_col + 8 - (c.loc.l_col mod 8) ; 
                        l_offset = c.loc.l_offset + 1 } 

let adj (buf : string) (thing : 'a) : ('a * span) = 
  let c = the_context in 
  let start_loc = c.loc in
  let len = (String.length buf) in 
  let end_loc = { c.loc with l_col = c.loc.l_col + len - 1} in 
  c.loc <- { c.loc with l_col = c.loc.l_col + len } ;
  let span = (start_loc, end_loc) in 
  (thing, span) 

let strip str = 
  let len = String.length str in
  String.sub str 1 (len - 2) 

let input_error sort_msg msg =
  let c = the_context in 
  let near_text = Lexing.lexeme c.lexbuf in 
  Printf.printf  
  "\n[%s] %s %s at line %d column %d\nNear Text: %s\n\t%s\n"
    c.filename 
    sort_msg
    (if c.warn_only then "WARNING" else "ERROR") 
    c.loc.l_line c.loc.l_col near_text msg ;
  raise Parsing.Parse_error

let lex_error msg = input_error "LEXER" msg 
let parse_error msg = input_error "PARSE" msg 
